//#define DEBUG
/*
Copyright (c) 2015, Raspberry Pi Foundation
Copyright (c) 2015, Dave Stevenson
Copyright (c) 2020-2021, Soho Enterprise Ltd.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the copyright holder nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <bcm_host.h>

static const char __PROGRAM_VERSION__[] = "@(#)i2c.c Ver. 1.0030 (2021/10/04 10:46:59)";

#define I2C_SLAVE_FORCE 0x0706

static int  I2C_ADDRESS = 0x10;

static char i2c_device_name[] = "/dev/i2c-0";

static int  Width;
static int  Height;
static int  BufSiz;
static int  DataLanes;
static int  RawBit;

#define NAME_LEN_MAX  32
#define TABLE_MAX 128

struct val_table {
    char name[NAME_LEN_MAX];
    int  value;
} table[TABLE_MAX];
int table_n = 0;

/*----------------------------------------------------------------------------*
 *
 *----------------------------------------------------------------------------*/
static char *
strip(char *s)
{
    char    *p;

    if ((p = strchr(s, 0x0A)) != NULL) {
        *p = '\0';
        for (--p; isspace(*p); p--) {
            *p = '\0';
        }
    }

    for (p = s; isspace(*p); p++) {
        ;
    }
    if (*p == '@') {
        return p;
    }

    if ((p = strstr(s, "//")) != NULL) {
        *p = '\0';
        for (--p; isspace(*p); p--) {
            *p = '\0';
        }
    }
    if ((p = strchr(s, '#')) != NULL) {
        *p = '\0';
        for (--p; isspace(*p); p--) {
            *p = '\0';
        }
    }
    if (strlen(s) == 0) {
        return  NULL;
    }
    return  s;
}

/*----------------------------------------------------------------------------*
 *
 *----------------------------------------------------------------------------*/
static char *
gettok(char *s, char *token)
{
    char    *p;

    if (s == NULL || *s == '\0')
        return NULL;

    while (isspace(*s)) {
        s++;
    }

    *token = '\0';
    for (p = token; *s && !isspace(*s); s++) {
        *p++ = *s;
    }
    *p = '\0';
    return s;
}

/*----------------------------------------------------------------------------*
 *
 *----------------------------------------------------------------------------*/
static int
atox(char *s)
{
    int    n;

    for (n = 0; isxdigit(*s); s++) {
        if (isdigit(*s)) {
            n = n * 16 + *s - '0';
        } else {
            n = n * 16 + toupper(*s) - 'A' + 10;
        }
    }
    return n;
}

/*----------------------------------------------------------------------------*
 *
 *----------------------------------------------------------------------------*/
static int
todigit(char *s)
{
    char    *p;
    int     n, i;

    if (*s == '$') {
        for (i = 0; i < table_n; i++) {
            if (strcmp(table[i].name, s) == 0) {
                n = table[i].value;
                break;
            }
        }
        if (i == table_n) {
            fprintf(stderr, "%s not found\n", s);
            exit(1);
        }
    } else {
        if (strncasecmp(s, "0x", 2) == 0) {
            n = atox(s+2);
        } else {
            for (p = s; *p != '\0'; p++) {
                if (!isdigit(*p)) {
                    fprintf(stderr, "%s is not digit\n", s);
                    exit(1);
                }
            }
            n = atoi(s);
        }
    }

    return  n;
}

/*----------------------------------------------------------------------------*
 *
 *----------------------------------------------------------------------------*/
int
SetReg(int reg16, int len, int data)
{
    int fd;
    unsigned char msg[6];

    if ((fd = open(i2c_device_name, O_RDWR)) == 0) {
        fprintf(stderr, "SetReg: error@open %s\n", strerror(errno));
        return  -1;
    }
    if (ioctl(fd, I2C_SLAVE_FORCE, I2C_ADDRESS) < 0) {
        fprintf(stderr, "SetReg: error@ioctl %s\n", strerror(errno));
        return  -1;
    }

#if 0
    if (len == 1) {
      msg[0] = (unsigned char)((reg16 + 0) >> 8);
      msg[1] = (unsigned char)((reg16 + 0) & 0xFF);
      msg[2] = (unsigned char)((data     ) & 0xFF);
      write(fd, msg, 3);
    }

    if (len == 2) {
        msg[0] = (unsigned char)((reg16 + 0) >> 8);
        msg[1] = (unsigned char)((reg16 + 0) & 0xFF);
        msg[2] = (unsigned char)((data >> 8) & 0xFF);
        write(fd, msg, 3);

        msg[0] = (unsigned char)((reg16 + 1) >> 8);
        msg[1] = (unsigned char)((reg16 + 1) & 0xFF);
        msg[2] = (unsigned char)((data     ) & 0xFF);
        write(fd, msg, 3);
    }
#else
    int     i, k;

//printf("## 0x%04x %04x\n", reg16, data);
    for (i = 0; i < len; i++) {
        k = (len - i - 1) * 8;
        msg[0] = (unsigned char)((reg16 + i) >> 8);
        msg[1] = (unsigned char)((reg16 + i) & 0xFF);
        msg[2] = (unsigned char)((data >> k) & 0xFF);
//printf(">> 0x%02x%02x 0x%02X\n", msg[0], msg[1], msg[2]);
        write(fd, msg, 3);
    }
#endif

    close(fd);

    return  0;
}

/*----------------------------------------------------------------------------*
 *
 *----------------------------------------------------------------------------*/
int
GetReg(int reg16, int len)
{
    int     fd;
    int     data;
    unsigned char rdata[2];
    unsigned char msg[6];

    if ((fd = open(i2c_device_name, O_RDWR)) == 0) {
        fprintf(stderr, "GetReg: error@open %s\n", strerror(errno));
        return  -1;
    }
    if (ioctl(fd, I2C_SLAVE_FORCE, I2C_ADDRESS) < 0) {
        fprintf(stderr, "GetReg: error@ioctl %s\n", strerror(errno));
        return  -1;
    }

    msg[0] = (unsigned char)((reg16 + 0) >> 8);
    msg[1] = (unsigned char)((reg16 + 0) & 0xFF);
    write(fd, msg, 2);
int n;
    if ((n = read(fd, rdata, len)) != len) {
        fprintf(stderr, "GetReg: error@read %s\n", strerror(errno));
        fprintf(stderr, "I2C_ADDRESS: %x\n", I2C_ADDRESS);
        fprintf(stderr, "len: len=%d  n=%d\n", len, n);
        exit(1);
        return  -1;
    }
    if (len == 1) {
        data = rdata[0];
    } else if (len == 2) {
        data = rdata[0] * 0x100 + rdata[1];
    } else {
        fprintf(stderr, "GetReg: len=%d No support (len > 2)\n", len);
        return  -1;
    }

    close(fd);

    return  data;
}

/*----------------------------------------------------------------------------*
 *
 *----------------------------------------------------------------------------*/
static void
Set1Byte(int reg, int val)
{
#ifdef DEBUG
    int     n, v;

    n = SetReg(reg, 1, val);
    v = GetReg(reg, 1);
    printf("(%2d) %04X: %02X %02X  %s\n", n, reg, val, v, (val==v)?"OK":"NG");
#else
    (void)SetReg(reg, 1, val);
#endif
}

/*----------------------------------------------------------------------------*
 *
 *----------------------------------------------------------------------------*/
static void
Set2Byte(int reg, int val)
{
#ifdef DEBUG
    int     n, v;

    n = SetReg(reg, 2, val);
    v = GetReg(reg, 2);
    printf("(%2d) %04X: %04X %04X  %s\n", n, reg, val, v, (val==v)?"OK":"NG");
#else
    (void)SetReg(reg, 2, val);
#endif
}

/*----------------------------------------------------------------------------*
 *
 *----------------------------------------------------------------------------*/
int
i2c(char *fname)
{
    FILE    *fp;
    char    buf[BUFSIZ], *p, *q;
    char    addr[BUFSIZ], data[BUFSIZ];
    char    name[BUFSIZ], value[BUFSIZ];
    char    word[BUFSIZ];
    int     i, n;
    int     reg, val;

    if ((fp = fopen(fname, "r")) == NULL) {
        fprintf(stderr, "can't open %s.\n", fname);
        return  -1;
    }

    table_n = 0;
    for (i = 0; fgets(buf, BUFSIZ, fp) != NULL; i++) {
        if ((p = strip(buf)) != NULL) {
            if (*p == '@') {
                printf("%s\n", p);
                continue;
            }
            if ((q = strchr(p, '=')) != NULL) {
                p = gettok(p  , name);
                p = gettok(q+1, value);
#ifdef  DEBUG
                printf("xxxxx %s = %s\n", name, value);
#endif
                if (strlen(name) + 1 >= NAME_LEN_MAX) {
                    fprintf(stderr, "over name length max(%d).\n", NAME_LEN_MAX);
                    exit(1);
                }
                strcpy(table[table_n].name, name);
                table[table_n].value = todigit(value);
                table_n++;
                if (table_n >= TABLE_MAX) {
                    fprintf(stderr, "over table max(%d).\n", TABLE_MAX);
                    exit(1);
                }
            } else {
                *word = '\0';
                p = gettok(p, addr);
                p = gettok(p, data);  val = todigit(data);
                p = gettok(p, word);

                if (strcasecmp(addr, "I2C_ADDRESS") == 0) {
                    I2C_ADDRESS  = val;
                } else if (strcasecmp(addr, "WIDTH") == 0) {
                    Width  = val;
                } else if (strcasecmp(addr, "HEIGHT") == 0) {
                    Height = val;
                } else if (strcasecmp(addr, "BUFSIZ") == 0) {
                    BufSiz = val;
                } else if (strcasecmp(addr, "DATA_LANES") == 0) {
                    DataLanes  = val;
                } else if (strcasecmp(addr, "RAWBIT") == 0) {
                    RawBit  = val;
                } else if (strcasecmp(addr, "WAIT") == 0) {
                    usleep(val * 1000);
                } else if (strcasecmp(addr, "DELAY") == 0) {
                    usleep(val * 1000);
                } else {
                    reg = todigit(addr);
#ifdef  DEBUG
                    printf("%3d : 0x%04X 0x%02X\n", i, reg, val);
#endif
                    if (*word == 'w' || *word == 'W') {
//                        printf("XXXX WORD DATA %04X %04X(%d)\n", todigit(addr), todigit(data), todigit(data));
                        Set2Byte(reg, val);
                    } else {
                        Set1Byte(reg, val);
                    }
                }
            }
        }
    }

    fclose(fp);

#ifdef  DEBUG
    for (i = 0; i < table_n; i++) {
        printf("%d: %8s = 0x%02X\n", i, table[i].name, table[i].value);
    }
#endif

    return  0;
}

/*----------------------------------------------------------------------------*
 *
 *----------------------------------------------------------------------------*/
int
check_lanes(char *fname)
{
    FILE    *fp;
    char    buf[BUFSIZ];
    char    *p, *q;
    char    addr[BUFSIZ], data[BUFSIZ];

    if ((fp = fopen(fname, "r")) == NULL) {
        fprintf(stderr, "can't open %s\n", fname);
        return  -1;
    }

    DataLanes = 2;
    while (fgets(buf, BUFSIZ, fp) != NULL) {
        if ((p = strip(buf)) != NULL) {
            p = gettok(p, addr);
            p = gettok(p, data);
            if (strcmp(addr, "DATA_LANES") == 0) {
                DataLanes = todigit(data);
                break;
            }
        }
    }
    fclose(fp);

    return  DataLanes;
}

/*----------------------------------------------------------------------------*
 *
 *----------------------------------------------------------------------------*/
int
check_rawbit(char *fname)
{
    FILE    *fp;
    char    buf[BUFSIZ];
    char    *p, *q;
    char    addr[BUFSIZ], data[BUFSIZ];

    if ((fp = fopen(fname, "r")) == NULL) {
        fprintf(stderr, "can't open %s\n", fname);
        return  -1;
    }

    RawBit = 10;
    while (fgets(buf, BUFSIZ, fp) != NULL) {
        if ((p = strip(buf)) != NULL) {
            p = gettok(p, addr);
            p = gettok(p, data);
            if (strcmp(addr, "RAWBIT") == 0) {
                RawBit = todigit(data);
                break;
            }
        }
    }
    fclose(fp);

    return  RawBit;
}

/*----------------------------------------------------------------------------*
 *
 *----------------------------------------------------------------------------*/
void
DispMode()
{
    printf("%s\n", __PROGRAM_VERSION__);
}

/*----------------------------------------------------------------------------*
 *
 *----------------------------------------------------------------------------*/
void
SetFrameLength(int val)
{
     fprintf(stderr, "SetFrameLength : no support. use SetReg.\n");
}

/*----------------------------------------------------------------------------*
 *
 *----------------------------------------------------------------------------*/
void
SetExposure(int val)
{
     fprintf(stderr, "SetExposure : no support. use SetReg.\n");
}

/*----------------------------------------------------------------------------*
 *
 *----------------------------------------------------------------------------*/
void
SetAnaGain(int val)
{
     fprintf(stderr, "SetAnaGain : no support. use SetReg.\n");
}

/*----------------------------------------------------------------------------*
 *
 *----------------------------------------------------------------------------*/
void
SetDigGain(int val)
{
     fprintf(stderr, "SetDigGain : no support. use SetReg.\n");
}

/*----------------------------------------------------------------------------*
 *
 *----------------------------------------------------------------------------*/
void
SetOrientation(int val)
{
     fprintf(stderr, "SetOrientation : no support. use SetReg.\n");
}

/*----------------------------------------------------------------------------*
 *
 *----------------------------------------------------------------------------*/
int
InitSensor(char *fname, int *width, int *height, int *bufsiz)
{
    Width     = 0;
    Height    = 0;
    BufSiz    = 0;
    DataLanes = 2;

    if (i2c(fname) < 0) { // Width,Height,Buflen will be changed
        return  -1;
    }

    if (Width == 0 || Height == 0) {
        fprintf(stderr, "can't set WIDTH and/or HEIGHT\n");
        return  -1;
    }
    *width  = Width;
    *height = Height;
    *bufsiz = BufSiz;

    printf("--- %s ---\n", fname);
    printf("RAWBIT     : %4d\n", RawBit);
    printf("DATA_LANES : %4d\n", DataLanes);
    printf("I2C_ADDRESS: 0x%02X\n", I2C_ADDRESS);
    printf("WIDTH      : %4d\n", Width );
    printf("HEIGHT     : %4d\n", Height);
//    printf("BUFSIZ     : %d  (width->%d)\n",  BufSiz, BufSiz/Height);
    printf("BUFSIZ     : %d\n",  BufSiz);
    printf("----------\n");

    return  0;
}

/*----------------------------------------------------------------------------*
 *
 *----------------------------------------------------------------------------*/
void
StopSensor()
{
    Set1Byte(0x0100, 0x00);    // stop
}

